@@ -40,64 +40,109 @@
 #include <libgnomevfs/gnome-vfs-mime.h>
 #include <gconf/gconf-client.h>
 #include "file-utils.h"
-#include "utf8-fnmatch.h"
 
 
 #define BUF_SIZE 4096
 #define MAX_PATTERNS 128
+#define MAX_SYMLINKS_FOLLOWED 32
 
 
 gboolean
-path_is_file (const gchar *path)
+path_exists (const gchar *path)
 {
-	GnomeVFSFileInfo *info;
-	GnomeVFSResult result;
-	gboolean is_file;
-	gchar *escaped;
+        GnomeVFSFileInfo *info;
+        GnomeVFSResult result;
+        gboolean exists;
+        gchar *escaped;
 
-	if (! path || ! *path) return FALSE; 
+        if (! path || ! *path) return FALSE;
 
-	info = gnome_vfs_file_info_new ();
-	escaped = gnome_vfs_escape_path_string (path);
-	result = gnome_vfs_get_file_info (escaped, 
-					  info, 
-					  (GNOME_VFS_FILE_INFO_DEFAULT 
-					   | GNOME_VFS_FILE_INFO_FOLLOW_LINKS));
-	is_file = FALSE;
-	if (result == GNOME_VFS_OK)		
-		is_file = (info->type == GNOME_VFS_FILE_TYPE_REGULAR);
-	
-	g_free (escaped);
-	gnome_vfs_file_info_unref (info);
+        info = gnome_vfs_file_info_new ();
+        escaped = gnome_vfs_escape_path_string (path);
+        result = gnome_vfs_get_file_info (escaped,
+                                          info,
+                                          (GNOME_VFS_FILE_INFO_DEFAULT
+                                           | GNOME_VFS_FILE_INFO_FOLLOW_LINKS));
 
-	return is_file;
+        exists = (result == GNOME_VFS_OK);
+
+        g_free (escaped);
+        gnome_vfs_file_info_unref (info);
+
+        return exists;
 }
 
 
 gboolean
-path_is_dir (const gchar *path)
+uri_exists (const char  *uri)
 {
-	GnomeVFSFileInfo *info;
-	GnomeVFSResult result;
-	gboolean is_dir;
-	gchar *escaped;
+        GnomeVFSFileInfo *info;
+        GnomeVFSResult    result;
+        gboolean          exists;
 
-	if (! path || ! *path) return FALSE; 
+        if (uri == NULL) 
+                return FALSE;
 
-	info = gnome_vfs_file_info_new ();
-	escaped = gnome_vfs_escape_path_string (path);
-	result = gnome_vfs_get_file_info (escaped, 
-					  info, 
-					  (GNOME_VFS_FILE_INFO_DEFAULT 
-					   | GNOME_VFS_FILE_INFO_FOLLOW_LINKS));
-	is_dir = FALSE;
-	if (result == GNOME_VFS_OK)
-		is_dir = (info->type == GNOME_VFS_FILE_TYPE_DIRECTORY);
-	
-	g_free (escaped);
-	gnome_vfs_file_info_unref (info);
+        info = gnome_vfs_file_info_new ();
+        result = gnome_vfs_get_file_info (uri,
+                                          info,
+                                          GNOME_VFS_FILE_INFO_DEFAULT);
 
-	return is_dir;
+        exists = (result == GNOME_VFS_OK);
+
+        gnome_vfs_file_info_unref (info);
+
+        return exists;  
+}
+
+
+gboolean
+uri_is_file (const char *uri)
+{
+        GnomeVFSFileInfo *info;
+        GnomeVFSResult    result;
+        gboolean          is_file;
+
+        if (! uri || ! *uri)
+                return FALSE;
+
+        info = gnome_vfs_file_info_new ();
+        result = gnome_vfs_get_file_info (uri,
+                                          info,
+                                          (GNOME_VFS_FILE_INFO_DEFAULT
+                                           | GNOME_VFS_FILE_INFO_FOLLOW_LINKS));
+        is_file = FALSE;
+        if (result == GNOME_VFS_OK)
+                is_file = (info->type == GNOME_VFS_FILE_TYPE_REGULAR);
+
+        gnome_vfs_file_info_unref (info);
+
+        return is_file;
+}
+
+
+gboolean
+uri_is_dir (const char *uri)
+{
+        GnomeVFSFileInfo *info;
+        GnomeVFSResult    result;
+        gboolean          is_dir;
+
+        if (! uri || ! *uri)
+                return FALSE;
+
+        info = gnome_vfs_file_info_new ();
+        result = gnome_vfs_get_file_info (uri,
+                                          info,
+                                          (GNOME_VFS_FILE_INFO_DEFAULT
+                                           | GNOME_VFS_FILE_INFO_FOLLOW_LINKS));
+        is_dir = FALSE;
+        if (result == GNOME_VFS_OK)
+                is_dir = (info->type == GNOME_VFS_FILE_TYPE_DIRECTORY);
+
+        gnome_vfs_file_info_unref (info);
+
+        return is_dir;
 }
 
 
@@ -155,7 +200,7 @@ get_file_size (const gchar *path)
 	if (! path || ! *path) return 0; 
 
 	info = gnome_vfs_file_info_new ();
-	escaped = gnome_vfs_escape_path_string (path);
+	escaped = gnome_vfs_escape_host_and_path_string (path);
 	result = gnome_vfs_get_file_info (escaped, 
 					  info,
 					  (GNOME_VFS_FILE_INFO_DEFAULT 
@@ -182,7 +227,7 @@ get_file_mtime (const gchar *path)
 	if (! path || ! *path) return 0; 
 
 	info = gnome_vfs_file_info_new ();
-	escaped = gnome_vfs_escape_path_string (path);
+	escaped = gnome_vfs_escape_host_and_path_string (path);
 	result = gnome_vfs_get_file_info (escaped, 
 					  info, 
 					  (GNOME_VFS_FILE_INFO_DEFAULT 
@@ -209,7 +254,7 @@ get_file_ctime (const gchar *path)
 	if (! path || ! *path) return 0; 
 
 	info = gnome_vfs_file_info_new ();
-	escaped = gnome_vfs_escape_path_string (path);
+	escaped = gnome_vfs_escape_host_and_path_string (path);
 	result = gnome_vfs_get_file_info (escaped, 
 					  info, 
 					  (GNOME_VFS_FILE_INFO_DEFAULT 
@@ -226,50 +271,27 @@ get_file_ctime (const gchar *path)
 
 
 gboolean 
-file_copy (const gchar *from, 
-	   const gchar *to)
+file_copy (const char *source_uri, 
+	   const char *dest_uri)
 {
-	FILE *fin, *fout;
-	gchar buf[BUF_SIZE];
-	gchar *dest_dir;
-	gint  n;
-
-	if (strcmp (from, to) == 0) {
-		g_warning ("cannot copy file %s: source and destination are the same\n", from);
-		return FALSE;
-	}
-
-	fin = fopen (from, "rb");
-	if (! fin) 
-		return FALSE;
-
-	dest_dir = remove_level_from_path (to);
-	if (! ensure_dir_exists (dest_dir, 0755)) {
-		g_free (dest_dir);
-		fclose (fin);
-		return FALSE;
-	}
-
-	fout = fopen (to, "wb");
-	if (! fout) {
-		g_free (dest_dir);
-		fclose (fin);
-		return FALSE;
-	}
-
-	while ((n = fread (buf, sizeof (char), BUF_SIZE, fin)) != 0) 
-		if (fwrite (buf, sizeof (char), n, fout) != n) {
-			g_free (dest_dir);
-			fclose (fin);
-			fclose (fout);
-			return FALSE;
-		}
-
-	g_free (dest_dir);
-	fclose (fin);
-	fclose (fout);
-
-	return TRUE;
+        GnomeVFSURI    *real_source_uri;
+        GnomeVFSURI    *real_dest_uri;
+        GnomeVFSResult  result;
+                
+        real_source_uri = gnome_vfs_uri_new (source_uri);
+        real_dest_uri = gnome_vfs_uri_new (dest_uri);
+                
+        result = gnome_vfs_xfer_uri (real_source_uri, 
+        			     real_dest_uri,
+                                     GNOME_VFS_XFER_RECURSIVE, 
+                                     GNOME_VFS_XFER_ERROR_MODE_ABORT,
+                                     GNOME_VFS_XFER_OVERWRITE_MODE_REPLACE,
+                                     NULL, NULL);
+                
+        gnome_vfs_uri_unref (real_source_uri);
+        gnome_vfs_uri_unref (real_dest_uri);
+                
+        return result == GNOME_VFS_OK;
 }
 
 
@@ -377,40 +399,60 @@ remove_ending_separator (const gchar *path)
 }
 
 
-gboolean 
-ensure_dir_exists (const gchar *a_path,
+GnomeVFSResult
+ensure_dir_exists (const gchar *uri,
 		   mode_t       mode)
 {
-	if (! a_path) return FALSE;
+	char *path;
+	char *p;
+	char *scheme;
 
-	if (! path_is_dir (a_path)) {
-		gchar *path = g_strdup (a_path);
-		gchar *p = path;
+	if (uri == NULL) 
+		return GNOME_VFS_ERROR_BAD_PARAMETERS;
+	
+	if (uri_is_dir (uri))
+		return GNOME_VFS_OK;
 
-		while (*p != '\0') {
-			p++;
-			if ((*p == '/') || (*p == '\0')) {
-				gboolean end = TRUE;
+	path = g_strdup (uri);
+	p = path;
 
-				if (*p != '\0') {
-					*p = '\0';
-					end = FALSE;
-				}
+	scheme = gnome_vfs_get_uri_scheme (path);
+	if (scheme != NULL) {
+		p += strlen (scheme);
+		if (strncmp (p, "://", 3) == 0)
+			p += 3;
+	}
+	
+	while (*p != '\0') {
+		gboolean end;
 
-			if (! path_is_dir (path)) {
-					if (mkdir (path, mode) < 0) {
-						g_warning ("directory creation failed: %s.", path);
-						g_free (path);
-						return FALSE;
-					}
-				}
-				if (! end) *p = '/';
+		p++;
+		if ((*p != '/') && (*p != '\0')) 
+			continue;
+
+		end = TRUE;
+		if (*p != '\0') {
+			*p = '\0';
+			end = FALSE;
+		}
+			
+		if (! uri_is_dir (path)) {
+			GnomeVFSResult result;
+			result = gnome_vfs_make_directory (path, mode);
+			if (result != GNOME_VFS_OK) {
+				g_free (path);
+				return result;
 			}
 		}
-		g_free (path);
-	}
 
-	return TRUE;
+		if (! end) 
+			*p = '/';
+	}
+	
+	g_free (scheme);
+	g_free (path);
+	
+	return GNOME_VFS_OK;
 }
 
 
@@ -545,7 +587,7 @@ escape_str (const char *str,
 char*
 shell_escape (const char *filename)
 {
-	return escape_str (filename, "$\'`\"\\!?* ()[]&|@#:;");
+	return escape_str (filename, "$\'`\"\\!?* ()[]<>&|@#:;");
 }
 
 
@@ -664,31 +706,6 @@ application_get_command (const GnomeVFSMimeApplication *app)
 }
 
 
-gboolean
-match_patterns (char       **patterns, 
-		const char  *string,
-		int          flags)
-{
-	int i;
-	int result;
-       
-	if (patterns[0] == NULL)
-		return TRUE;
-	
-	if (string == NULL)
-		return FALSE;
-	
-	result = FNM_NOMATCH;
-	i = 0;
-	while ((result != 0) && (patterns[i] != NULL)) {
-		result = g_utf8_fnmatch (patterns[i], string, flags);
-		i++;
-	}
-
-	return (result == 0);
-}
-
-
 static const char *
 g_utf8_strstr (const char *haystack, const char *needle)
 {
@@ -837,7 +854,7 @@ get_dest_free_space (const char  *path)
 	GnomeVFSResult    result;
 	GnomeVFSFileSize  ret_val;
 
-	escaped = gnome_vfs_escape_path_string (path);
+	escaped = gnome_vfs_escape_host_and_path_string (path);
 	uri = gnome_vfs_uri_new (escaped);
 	g_free (escaped);
 
@@ -914,7 +931,7 @@ rmdir_recursive (const gchar *directory)
 	GList    *scan;
 	gboolean  error = FALSE;
 
-	if (! path_is_dir (directory)) 
+	if (! uri_is_dir (directory)) 
 		return FALSE;
 
 	path_list_new (directory, &files, &dirs);
@@ -1021,127 +1038,12 @@ get_temp_work_dir_name (void)
 					  "/goobox",
 					  getpid (),
 					  count++);
-	} while (path_is_file (result) && (try++ < MAX_TRIES));
+	} while (uri_is_file (result) && (try++ < MAX_TRIES));
 
 	return result;
 }
 
 
-/* file list utils */
-
-
-gboolean
-file_list__match_pattern (const char *line, 
-			  const char *pattern)
-{
-	const char *l = line, *p = pattern;
-
-	for (; (*p != 0) && (*l != 0); p++, l++) {
-		if (*p != '%') {
-			if (*p != *l)
-				return FALSE;
-		} else {
-			p++;
-			switch (*p) {
-			case 'a':
-				break;
-			case 'n':
-				if (!isdigit (*l)) 
-					return FALSE;
-				break;
-			case 'c':
-				if (!isalpha (*l)) 
-					return FALSE;
-				break;
-			default:
-				return FALSE;
-			}
-		}
-	}
-	
-	return (*p == 0);
-}
-
-
-int
-file_list__get_index_from_pattern (const char *line, 
-				   const char *pattern)
-{
-	int         line_l, pattern_l;
-	const char *l;
-
-	line_l = strlen (line);
-	pattern_l = strlen (pattern);
-
-	if ((pattern_l == 0) || (line_l == 0))
-		return -1;
-
-	for (l = line; *l != 0; l++) 
-		if (file_list__match_pattern (l, pattern))
-			return (l - line);
-	
-	return -1;
-}
-
-
-char*
-file_list__get_next_field (const char *line,
-			   int         start_from,
-			   int         field_n)
-{
-	const char *f_start, *f_end;
-	
-	line = line + start_from;
-	
-	f_start = line;
-	while ((*f_start == ' ') && (*f_start != *line))
-		f_start++;
-	f_end = f_start;
-
-	while ((field_n > 0) && (*f_end != 0)) {
-		if (*f_end == ' ') {
-			field_n--;
-			if (field_n != 0) {
-				while ((*f_end == ' ') && (*f_end != *line))
-					f_end++;
-				f_start = f_end;
-			}
-		} else
-			f_end++;
-	}
-	
-	return g_strndup (f_start, f_end - f_start);
-}
-
-
-char*
-file_list__get_prev_field (const char *line,
-			   int         start_from,
-			   int         field_n)
-{
-	const char *f_start, *f_end;
-
-	f_start = line + start_from - 1;
-	while ((*f_start == ' ') && (*f_start != *line))
-		f_start--;
-	f_end = f_start;
-
-	while ((field_n > 0) && (*f_start != *line)) {
-		if (*f_start == ' ') {
-			field_n--;
-			if (field_n != 0) {
-				while ((*f_start == ' ') && (*f_start != *line))
-					f_start--;
-				f_end = f_start;
-			}
-		} else
-			f_start--;
-	}
-
-	return g_strndup (f_start + 1, f_end - f_start);
-}
-
-
 GnomeVFSURI *
 new_uri_from_path (const char *path)
 {
@@ -1164,3 +1066,518 @@ new_uri_from_path (const char *path)
 
 	return uri;
 }
+
+
+char *
+get_uri_from_local_path (const char *local_path)
+{
+        char *escaped;
+        char *uri;
+
+        escaped = escape_uri (local_path);
+        if (escaped[0] == '/') {
+                uri = g_strconcat ("file://", escaped, NULL);
+                g_free (escaped);
+        }
+        else
+                uri = escaped;
+
+        return uri;
+}
+
+
+const char *
+get_path_from_uri (const char *uri)
+{
+	const char *file_prefix = "file://";
+	int         file_prefix_l = strlen (file_prefix);
+	
+	if (strncmp (uri, file_prefix, file_prefix_l) == 0)
+		uri += file_prefix_l;
+	return uri;
+}
+
+
+static gboolean
+valid_filename_char (char c) 
+{
+	/* "$\'`\"\\!?* ()[]<>&|@#:;" */
+	return strchr ("/\\!?*:;'`\"", c) == NULL;
+}
+
+
+/* Remove special characters from a track title in order to make it a 
+ * valid filename. */
+char*
+tracktitle_to_filename (const char *trackname)
+{
+	char       *filename, *f, *f2;
+	const char *t;
+	gboolean    add_space;
+	char       *uri_name;
+	
+	if (trackname == NULL)
+		return NULL;
+
+	filename = g_new (char, strlen (trackname) + 1);
+
+	/* Substitute invalid characters with spaces. */
+	f = filename;
+	t = trackname;
+	while (*t != 0) {
+		gboolean invalid_char = FALSE;
+
+		while ((*t != 0) && ! valid_filename_char (*t)) {
+			invalid_char = TRUE;
+			t++;
+		}
+
+		if (invalid_char)
+			*f++ = ' ';
+		
+		*f = *t;
+
+		if (*t != 0) {
+			f++;
+			t++;
+		}
+	}
+	*f = 0;
+
+	/* Remove double spaces. */
+	add_space = FALSE;
+	for (f = f2 = filename; *f != 0; f++)
+		if (*f != ' ') {
+			if (add_space) {
+				*f2++ = ' ';
+				add_space = FALSE;
+			}
+			*f2 = *f;
+			f2++;
+		} else
+			add_space = TRUE;
+	*f2 = 0;
+
+	uri_name = gnome_vfs_escape_string (filename);
+	g_free (filename);
+
+	return uri_name;
+}
+
+
+gboolean
+is_program_in_path (const char *program_name)
+{
+	char *str;
+        int   result = FALSE;
+	
+        str = g_find_program_in_path (program_name);
+        if (str != NULL) {
+                g_free (str);
+                result = TRUE;
+        }
+	
+        return result;
+}
+
+
+char *
+xdg_user_dir_lookup (const char *type)
+{
+        FILE *file;
+        char *home_dir, *config_home, *config_file;
+        char buffer[512];
+        char *user_dir;
+        char *p, *d;
+        int len;
+        int relative;
+
+        home_dir = getenv ("HOME");
+        if (home_dir == NULL)
+                return strdup ("/tmp");
+
+        config_home = getenv ("XDG_CONFIG_HOME");
+        if (config_home == NULL || config_home[0] == 0) {
+                config_file = malloc (strlen (home_dir) + strlen ("/.config/user-dirs.dirs") + 1);
+                strcpy (config_file, home_dir);
+                strcat (config_file, "/.config/user-dirs.dirs");
+        } else {
+                config_file = malloc (strlen (config_home) + strlen ("/user-dirs.dirs") + 1);
+                strcpy (config_file, config_home);
+                strcat (config_file, "/user-dirs.dirs");
+        }
+
+        file = fopen (config_file, "r");
+        free (config_file);
+        if (file == NULL)
+                goto error;
+
+        user_dir = NULL;
+
+        while (fgets (buffer, sizeof (buffer), file)) {
+                /* Remove newline at end */
+                len = strlen (buffer);
+                if (len > 0 && buffer[len-1] == '\n')
+                buffer[len-1] = 0;
+      
+                p = buffer;
+                while (*p == ' ' || *p == '\t')
+                        p++;
+      
+                if (strncmp (p, "XDG_", 4) != 0)
+                        continue;
+                p += 4;
+
+                if (strncmp (p, type, strlen (type)) != 0)
+                        continue;
+                p += strlen (type);
+
+                if (strncmp (p, "_DIR", 4) != 0)
+                continue;
+                
+                p += 4;
+
+                while (*p == ' ' || *p == '\t')
+                        p++;
+
+                if (*p != '=')
+                        continue;
+                p++;
+      
+                while (*p == ' ' || *p == '\t')
+                        p++;
+
+                if (*p != '"')
+                        continue;
+                p++;
+      
+                relative = 0;
+                if (strncmp (p, "$HOME/", 6) == 0) {
+                        p += 6;
+                        relative = 1;
+                } else if (*p != '/')
+                        continue;
+                        if (relative) {
+                        user_dir = malloc (strlen (home_dir) + 1 + strlen (p) + 1);
+                        strcpy (user_dir, home_dir);
+                        strcat (user_dir, "/");
+                } else {
+                        user_dir = malloc (strlen (p) + 1);
+                        *user_dir = 0;
+                }
+      
+                d = user_dir + strlen (user_dir);
+                while (*p && *p != '"') {
+                        if ((*p == '\\') && (*(p+1) != 0))
+                                p++;
+                        *d++ = *p++;
+                }
+                *d = 0;
+        }  
+        
+        fclose (file);
+
+        if (user_dir)
+                return user_dir;
+
+error:
+         /* Special case desktop for historical compatibility */
+        if (strcmp (type, "DESKTOP") == 0) {
+                user_dir = malloc (strlen (home_dir) + strlen ("/Desktop") + 1);
+                strcpy (user_dir, home_dir);
+                strcat (user_dir, "/Desktop");
+                return user_dir;
+        } else
+                return strdup (home_dir);
+}
+
+
+char *
+get_uri_host (const char *uri)
+{
+	const char *idx;
+
+	idx = strstr (uri, "://");
+	if (idx == NULL)
+		return NULL;
+	idx = strstr (idx + 3, "/");
+	if (idx == NULL)
+		return NULL;
+	return g_strndup (uri, (idx - uri));
+}
+
+
+static char *
+build_uri (const char *s1,
+	   const char *s2)
+{
+	char *s = NULL;
+
+	if ((s1 == NULL) && (s2 == NULL))
+		return NULL;
+
+	if ((s1 == NULL) || (strcmp (s1, "") == 0))
+		return g_strdup (s2);
+	if ((s2 == NULL) || (strcmp (s2, "") == 0))
+		return g_strdup (s1);
+
+	if ((s1[strlen (s1) - 1] == GNOME_VFS_URI_PATH_CHR) || (s2[0] == GNOME_VFS_URI_PATH_CHR))
+		s = g_strconcat (s1, s2, NULL);
+	else
+		s = g_strconcat (s1, GNOME_VFS_URI_PATH_STR, s2, NULL);
+
+	return s;
+}
+
+
+const char *
+remove_host_from_uri (const char *uri)
+{
+	const char *idx;
+
+	idx = strstr (uri, "://");
+	if (idx == NULL)
+		return uri;
+	idx = strstr (idx + 3, "/");
+	if (idx == NULL)
+		return NULL;
+	return idx;
+}
+
+
+/*
+ * example 1) input : "/xxx/yyy/.."         output : "/xxx"
+ * example 2) input : "/xxx/yyy/../www"     output : "/xxx/www"
+ * example 3) input : "/xxx/../yyy/../www"  output : "/www"
+ * example 4) input : "/xxx/../yyy/../"     output : "/"
+ * example 5) input : "/xxx/./"             output : "/xxx"
+ * example 6) input : "/xxx/./yyy"          output : "/xxx/yyy"
+ *
+ * Note : PATH must be absolute.
+ */
+char *
+remove_special_dirs_from_path (const char *uri)
+{
+	const char  *path;
+	char       **pathv;
+	GList       *list = NULL;
+	int          i;
+	GString     *result_s;
+	char        *scheme;
+	char        *result;
+	int	     start_at;
+
+	scheme = get_uri_host (uri);
+
+	if ((scheme == NULL) && ! g_path_is_absolute (uri))
+		return g_strdup (uri);
+
+	path = remove_host_from_uri (uri);
+
+	if ((path == NULL) || (strstr (path, ".") == NULL))
+		return g_strdup (uri);
+
+	pathv = g_strsplit (path, "/", 0);
+
+	/* Trimmed uris might not start with a slash */
+	if (*path != '/')
+		start_at = 0;
+	else
+		start_at = 1;
+
+	/* Ignore first slash, if present. It will be re-added later. */
+	for (i = start_at; pathv[i] != NULL; i++) {
+		if (strcmp (pathv[i], ".") == 0) {
+			/* nothing to do. */
+		} else if (strcmp (pathv[i], "..") == 0) {
+			if (list == NULL) {
+				/* path error. */
+				g_strfreev (pathv);
+				return NULL;
+			}
+			list = g_list_delete_link (list, list);
+		} else
+			list = g_list_prepend (list, pathv[i]);
+	}
+
+	result_s = g_string_new (NULL);
+
+	/* re-insert URI scheme */
+	if (scheme != NULL) {
+		g_string_append (result_s, scheme);
+
+		if (start_at == 0)
+			/* delete trailing slash, because an extra one is added below */
+			g_string_truncate (result_s, result_s->len - 1);
+
+		g_free (scheme);
+	}
+
+	if (list == NULL)
+		g_string_append_c (result_s, '/');
+	else {
+		GList *scan;
+
+		list = g_list_reverse (list);
+		for (scan = list; scan; scan = scan->next) {
+			g_string_append_c (result_s, '/');
+			g_string_append (result_s, scan->data);
+		}
+	}
+	result = result_s->str;
+	g_string_free (result_s, FALSE);
+	g_strfreev (pathv);
+
+	return result;
+}
+
+
+static GnomeVFSResult
+resolve_symlinks (const char  *text_uri,
+		  const char  *relative_link,
+		  char       **resolved_text_uri,
+		  int          n_followed_symlinks)
+{
+	GnomeVFSResult     result = GNOME_VFS_OK;
+	char              *resolved_uri;
+	char 		  *uri;
+	char              *tmp;
+	GnomeVFSFileInfo  *info;
+	const char        *path;
+	char             **names;
+	int                i;
+
+	*resolved_text_uri = NULL;
+
+	if (text_uri == NULL)
+		return GNOME_VFS_OK;
+	if (*text_uri == '\0')
+		return GNOME_VFS_ERROR_INVALID_URI;
+
+	info = gnome_vfs_file_info_new ();
+
+	resolved_uri = get_uri_host (text_uri);
+	if (resolved_uri == NULL)
+		resolved_uri = g_strdup ("file://");
+
+	tmp = build_uri (text_uri, relative_link);
+	uri = remove_special_dirs_from_path (tmp);
+	g_free (tmp);
+
+	/* split the uri and resolve one name at a time. */
+
+	path = remove_host_from_uri (uri);
+	if (path == NULL) {
+		*resolved_text_uri = resolved_uri;
+		return GNOME_VFS_OK;
+	}
+
+	names = g_strsplit (path, GNOME_VFS_URI_PATH_STR, -1);
+	g_free (uri);
+
+	for (i = 0; (result == GNOME_VFS_OK) && (names[i] != NULL); i++) {
+		char  *try_uri;
+		char  *symlink;
+	    	char **symlink_names;
+	    	int    j;
+	    	char  *base_uri;
+
+		if (strcmp (names[i], "") == 0)
+			continue;
+
+		gnome_vfs_file_info_clear (info);
+
+		try_uri = g_strconcat (resolved_uri, GNOME_VFS_URI_PATH_STR, names[i], NULL);
+
+		result = gnome_vfs_get_file_info (try_uri, info, GNOME_VFS_FILE_INFO_DEFAULT);
+		if (result != GNOME_VFS_OK) {
+			g_free (try_uri);
+			break;
+		}
+
+		/* if names[i] isn't a symbolic link add it to the resolved uri and continue */
+
+		if (!((info->type == GNOME_VFS_FILE_TYPE_SYMBOLIC_LINK) &&
+		      (info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_SYMLINK_NAME))) {
+
+			g_free (resolved_uri);
+			resolved_uri = try_uri;
+
+			continue;
+		}
+
+		g_free (try_uri);
+
+		/* names[i] is a symbolic link */
+
+		n_followed_symlinks++;
+		if (n_followed_symlinks > MAX_SYMLINKS_FOLLOWED) {
+			result = GNOME_VFS_ERROR_TOO_MANY_LINKS;
+			break;
+		}
+
+		/* get the symlink escaping the value info->symlink_name */
+
+		symlink = g_strdup ("");
+		symlink_names = g_strsplit (info->symlink_name, GNOME_VFS_URI_PATH_STR, -1);
+		for (j = 0; symlink_names[j] != NULL; j++) {
+			char *symlink_name = symlink_names[j];
+	    		char *e_symlink_name;
+
+		    	if ((strcmp (symlink_name, "..") == 0) || (strcmp (symlink_name, ".") == 0))
+		    		e_symlink_name = g_strdup (symlink_name);
+		    	if (strcmp (symlink_name, "") == 0)
+		    		e_symlink_name = g_strdup (GNOME_VFS_URI_PATH_STR);
+		    	else
+		    		e_symlink_name = gnome_vfs_escape_string (symlink_name);
+
+		    	if (strcmp (symlink, "") == 0) {
+		    		g_free (symlink);
+		    		symlink = e_symlink_name;
+		    	}
+		    	else {
+		    		char *tmp;
+
+		   		tmp = build_uri (symlink, e_symlink_name);
+
+	    			g_free (symlink);
+	    			g_free (e_symlink_name);
+
+	    			symlink = tmp;
+	    		}
+	    	}
+	    	g_strfreev (symlink_names);
+
+		/* if the symlink is absolute reset the base uri, else use
+		 * the currently resolved uri as base. */
+
+	    	if (symlink[0] == GNOME_VFS_URI_PATH_CHR) {
+	    		g_free (resolved_uri);
+	    		base_uri = get_uri_host (text_uri);
+	    	} else
+	    		base_uri = resolved_uri;
+
+		/* resolve the new uri recursively */
+
+	    	result = resolve_symlinks (base_uri, symlink, &resolved_uri, n_followed_symlinks);
+
+	    	g_free (base_uri);
+		g_free (symlink);
+	}
+
+	g_strfreev (names);
+	gnome_vfs_file_info_unref (info);
+
+	if (result == GNOME_VFS_OK)
+		*resolved_text_uri = resolved_uri;
+
+	return result;
+}
+
+
+GnomeVFSResult
+resolve_all_symlinks (const char  *text_uri,
+		      char       **resolved_text_uri)
+{
+	return resolve_symlinks (text_uri, "", resolved_text_uri, 0);
+}
